"""Tests for the time_of_day timer."""

from datetime import datetime

import pytest

from astrality.timer import TimeOfDay, WorkDay


@pytest.fixture
def default_time_of_day_timer():
    """Return a default time_of_day timer object."""
    return TimeOfDay({'type': 'time_of_day'})


def test_processing_of_time_of_day_config(default_time_of_day_timer):
    """Test that the processing of the time_of_day config is correct."""
    assert 'sunday' not in default_time_of_day_timer.workdays
    assert 'monday' in default_time_of_day_timer.workdays

    assert default_time_of_day_timer.workdays['monday'].start.tm_hour == 9
    assert default_time_of_day_timer.workdays['monday'].start.tm_min == 0

    assert default_time_of_day_timer.workdays['friday'].end.tm_hour == 17
    assert default_time_of_day_timer.workdays['friday'].end.tm_min == 0


def test_current_period_of_time_of_day_timer(default_time_of_day_timer, freezer):
    work_monday = datetime(year=2018, month=2, day=12, hour=10)
    freezer.move_to(work_monday)
    assert default_time_of_day_timer.period() == 'on'

    monday_freetime = datetime(year=2018, month=2, day=12, hour=18)
    freezer.move_to(monday_freetime)
    assert default_time_of_day_timer.period() == 'off'

    saturday = datetime(year=2018, month=2, day=17, hour=10)
    freezer.move_to(saturday)
    assert default_time_of_day_timer.period() == 'off'

def test_time_until_next_period_for_time_of_day_timer(
    default_time_of_day_timer,
    freezer,
):
    work_monday = datetime(year=2018, month=2, day=12, hour=10)
    freezer.move_to(work_monday)
    assert default_time_of_day_timer.time_until_next_period().total_seconds() == 7*60*60 + 60

    monday_freetime = datetime(year=2018, month=2, day=12, hour=18)
    freezer.move_to(monday_freetime)
    assert default_time_of_day_timer.time_until_next_period().total_seconds() == 15*60*60 + 60

    saturday = datetime(year=2018, month=2, day=17, hour=10)
    freezer.move_to(saturday)
    assert default_time_of_day_timer.time_until_next_period().total_seconds() == 47*60*60 + 60



import os
import signal
import subprocess
import time

import pytest

from astrality.astrality import main


@pytest.mark.slow
def test_termination_of_main_process():
    astrality_process = subprocess.Popen(
        ['./bin/astrality'],
        stdout=subprocess.PIPE,
        preexec_fn=os.setsid,
    )

    time.sleep(2)
    os.killpg(os.getpgid(astrality_process.pid), signal.SIGTERM)

    astrality_process.wait()
    assert astrality_process.returncode == 0


@pytest.mark.slow
def test_interrupt_of_main_process():
    astrality_process = subprocess.Popen(
        ['./bin/astrality'],
        stdout=subprocess.PIPE,
        preexec_fn=os.setsid,
    )

    time.sleep(2)
    os.killpg(os.getpgid(astrality_process.pid), signal.SIGINT)

    astrality_process.wait()
    assert astrality_process.returncode == 0

@pytest.mark.slow
def test_invocation_of_main_process():
    main(test=True)



from math import inf
from numbers import Number
from typing import (
    Any,
    Dict,
    ItemsView,
    KeysView,
    ValuesView,
    Optional,
    Tuple,
    Union,
)


Real = Union[int, float]
Key = Union[str, Real]
Value = Any


class Resolver:
    """
    Dictionary-like object whith integer index resolution.

    An example of its functionality:
    replacements = Resolver({
    'colors': {1: 'CACCFD', 2: 'BACBEB'}
    })
    replacements['colors'][1]
    >>> 'CACCFD'
    replacements['colors'][2]
    >>> 'BACBEB'
    replacements['colors'][3]
    >>> 'BACBEB'
    """
    _dict: Union['Resolver', Dict[Key, Value]]

    def __init__(
        self,
        content: Optional[Union['Resolver', dict]] = None,
    ) -> None:
        """
        Initialize a Resolver instance from another Resolver or dictionary.

        If not given an argument, an empty Resolver object is initialized.
        """
        self._dict = {}
        self._max_key: Real = float(-inf)

        if isinstance(content, (Resolver, dict,)):
            self.update(content)
        elif content is not None:
            raise ValueError('Resolver initialized with wrong argument type.')

        # Determine the greatest number index inserted
        for key in self._dict.keys():
            if isinstance(key, Number):
                self._max_key = max(key, self._max_key)

    def __eq__(self, other) -> bool:
        """Check if content is identical to other Resolver or dictionary."""
        if isinstance(other, Resolver):
            return self._dict == other._dict
        elif isinstance(other, dict):
            return self._dict == other
        else:
            raise RuntimeError(
                f'Resolver comparison with unknown type "{type(other)}"',
            )

    def __setitem__(self, key: Key, value: Value) -> None:
        """Insert `value` into the `key` index."""
        if isinstance(key, Number):
            self._max_key = max(key, self._max_key)

        if isinstance(value, dict):
            # Insterted dictionaries are cast to Resolver instances
            self._dict[key] = Resolver(value)
        else:
            self._dict[key] = value

    def __getitem__(self, key: Key) -> Value:
        """
        Get item inserted into `key` index, with integer index resolution.

        Here "integer index resolution" means that if you try to retrieve
        non-existent integer index 2, it will retrieve the greatest available
        integer indexed value instead.
        """
        try:
            # Return excact hit if present
            return self._dict[key]
        except KeyError as key_error:
            # The key is not present. See if we can resolve the use of another
            # one through integer key priority.
            if self._max_key > -inf:
                # Another integer key has been inserted earlier
                if isinstance(key, Number):
                    # We can return the max integer key previously inserted
                    return self._dict[self._max_key]
                else:
                    # Throwing any other ValueErrors just in case
                    raise
            else:
                raise KeyError(f'Integer index "{key}" is non-existent and had '
                               'no lower index to be substituted for')


    def get(self, key: Key, defualt=None) -> Value:
        """Get value from index with fallback value `default`."""
        try:
            return self.__getitem__(key)
        except KeyError:
            return defualt

    def __repr__(self) -> str:
        """Return human-readable representation of Resolver object."""
        return f'Resolver({self._dict.__repr__()})'

    def __str__(self) -> str:
        """Return string representation of Resolver object."""
        return f'Resolver({self._dict.__str__()})'

    def __len__(self) -> int:
        """Return the number of key inserted into the Resolver object."""
        return self._dict.__len__()

    def __contains__(self, key: Key) -> bool:
        """Return true if `key` is inserted into Resolver object."""
        return self._dict.__contains__(key)

    def items(self) -> ItemsView[Key, Value]:
        """Return all key, value pairs of the Resolver object."""
        return self._dict.items()

    def keys(self) -> KeysView[Key]:
        """Return all keys which have been inserted into the Resolver object."""
        return self._dict.keys()

    def values(self) -> ValuesView[Value]:
        """Return all values inserted into the Resolver object."""
        return self._dict.values()

    def update(self, other: Union['Resolver', dict]) -> None:
        """Overwrite all items from other onto the Resolver object."""
        for key, value in other.items():
            self.__setitem__(key, value)



"""Module implementing user configured custom functionality."""

from collections import namedtuple
from datetime import timedelta
import logging
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Tuple

from astrality import compiler
from astrality.compiler import context
from astrality.config import ApplicationConfig, insert_into
from astrality.filewatcher import DirectoryWatcher
from astrality.timer import Timer, timer_factory
from astrality.utils import run_shell

ModuleConfig = Dict[str, Any]
ContextSectionImport = namedtuple(
    'ContextSectionImport',
    ['into_section', 'from_section', 'from_config_file'],
)
logger = logging.getLogger('astrality')


class Module:
    """
    Class for determining module actions.

    A Module instance should only return actions to be performed, but never
    perform them. That is the responsibility of a ModuleManager instance.

    A module can define a set of commands to be run on astrality startup, and
    exit, in addition to every time a given type of period changes.

    Commands are run in the users shell, and can use the following placeholders:
    - {period}: The period specified by the timer instance.
    - {name_of_template}: The path to the compiled template specified in the
                          'target' option, or if not specified, a created
                          temporary file.
    """

    def __init__(
        self,
        module_config: ModuleConfig,
        config_directory: Path,
        temp_directory: Path,
    ) -> None:
        """
        Initialize Module object with a section from a config dictionary.

        Section name must be [module/*], where * is the module name.
        In addition, the enabled option must be set to "true", or not set at
        all.

        module_config example:
        {'module/name':
            'enabled': True,
            'timer': {'type': 'weekday'},
            'on_startup': {'run': ['echo weekday is {period}']},
        }
        """
        # Can only initialize one module at a time
        assert len(module_config) == 1

        section = next(iter(module_config.keys()))
        self.name: str = section.split('/')[1]

        self.module_config = module_config[section]
        self.config_directory = config_directory
        self.temp_directory = temp_directory

        # Use static timer if no timer is specified
        self.timer: Timer = timer_factory(
            self.module_config.get('timer', {'type': 'static'}),
        )

        # Find and prepare templates and compilation targets
        self._prepare_templates()

    def _prepare_templates(self) -> None:
        """Determine template sources and compilation targets."""
        # config['templates'] is a dictionary with keys naming each template,
        # and values containing another dictionary containing the mandatory
        # key `source`, and the optional key `target`.
        templates: Dict[str, Dict[str, str]] = self.module_config.get(
            'templates',
            {},
        )
        self.templates: Dict[str, Dict[str, Path]] = {}

        for name, template in templates.items():
            source = self.expand_path(Path(template['source']))

            if not source.is_file():
                logger.error(\
                    f'[module/{self.name}] Template "{name}": source "{source}"'
                    ' does not exist. Skipping compilation of this file.'
                )
                continue

            config_target = template.get('target')
            if config_target:
                target = self.expand_path(Path(config_target))
            else:
                target = self.create_temp_file()

            self.templates[name] = {'source': source, 'target': target}

    def create_temp_file(self) -> Path:
        """Create a temp file used as a compilation target, returning its path."""

        temp_file = NamedTemporaryFile(  # type: ignore
            prefix=self.name + '-',
            dir=self.temp_directory,
        )

        # NB: These temporary files need to be persisted during the entirity of
        # the scripts runtime, since the files are deleted when they go out of
        # scope.
        if not hasattr(self, 'temp_files'):
            self.temp_files = [temp_file]
        else:
            self.temp_files.append(temp_file)

        return Path(temp_file.name)

    def expand_path(self, path: Path) -> Path:
        """
        Return an absolute path from a (possibly) relative path.

        Relative paths are relative to $ASTRALITY_CONFIG_HOME, and ~ is
        expanded to the home directory of $USER.
        """

        path = Path.expanduser(path)

        if not path.is_absolute():
            path = Path(
                self.config_directory,
                path,
            )

        return path

    def startup_commands(self) -> Tuple[str, ...]:
        """Return commands to be run on Module instance startup."""
        startup_commands: List[str] = self.module_config.get(
            'on_startup',
            {},
        ).get('run', [])

        if len(startup_commands) == 0:
            logger.debug(f'[module/{self.name}] No startup command specified.')
            return ()
        else:
            return tuple(
                self.interpolate_string(command)
                for command
                in startup_commands
            )

    def period_change_commands(self) -> Tuple[str, ...]:
        """Commands to be run when self.timer period changes."""
        period_change_commands: List[str] = self.module_config.get(
            'on_period_change',
            {},
        ).get('run', [])

        if len(period_change_commands) == 0:
            logger.debug(f'[module/{self.name}] No period change command specified.')
            return ()
        else:
            return tuple(
                self.interpolate_string(command)
                for command
                in period_change_commands
            )

    def exit_commands(self) -> Tuple[str, ...]:
        """Commands to be run on Module instance shutdown."""
        exit_commands: List[str] = self.module_config.get(
            'on_exit',
            {},
        ).get('run', [])

        if len(exit_commands) == 0:
            logger.debug(f'[module/{self.name}] No exit command specified.')
            return ()
        else:
            return tuple(
                self.interpolate_string(command)
                for command
                in exit_commands
            )

    def modified_commands(self, template_name: str) -> Tuple[str, ...]:
        """Commands to be run when a module template is modified."""
        modified_commands: List[str] = self.module_config.get(
            'on_modified',
            {},
        ).get(template_name, {}).get('run', [])

        if len(modified_commands) == 0:
            logger.debug(f'[module/{self.name}] No modified command specified.')
            return ()
        else:
            return tuple(
                self.interpolate_string(command)
                for command
                in modified_commands
            )


    def context_section_imports(
        self,
        trigger: str,
    ) -> Tuple[ContextSectionImport, ...]:
        """
        Return what to import into the global application_context.

        Trigger is one of 'on_startup', 'on_period_change', or 'on_exit'.
        This determines which section of the module is used to get the context
        import specification from.
        """
        assert trigger in ('on_startup', 'on_period_change', 'on_startup',)

        context_section_imports = []
        import_config = self.module_config.get(
            trigger,
            {},
        ).get('import_context', [])

        for context_import in import_config:
            # Insert placeholders
            from_file = self.interpolate_string(context_import['from_file'])
            from_section = self.interpolate_string(context_import['from_section'])

            # If no `to_section` is specified, use the same section as
            # `from_section`
            to_section = self.interpolate_string(
                context_import.get('to_section', from_section),
            )

            # Get the absolute path
            config_path = self.expand_path(Path(from_file))

            # Isert a ContextSectionImport tuple into the return value
            context_section_imports.append(
                ContextSectionImport(
                    into_section=to_section,
                    from_section=from_section,
                    from_config_file=config_path,
                )
            )

        return tuple(context_section_imports)


    def interpolate_string(self, string: str) -> str:
        """Replace all module placeholders in string."""
        string = string.format(
            # {period} -> current period defined by module timer
            period=self.timer.period(),
            # {name_of_template} -> string path to compiled template
            **{
                name: str(template['target'])
                for name, template
                in self.templates.items()
            }
        )
        return string

    @staticmethod
    def valid_class_section(section: ModuleConfig) -> bool:
        """Check if the given dict represents a valid enabled module."""

        if not len(section) == 1:
            raise RuntimeError(
                'Tried to check module section with dict '
                'which does not have exactly one item.',
            )

        try:
            module_name = next(iter(section.keys()))
            valid_module_name = module_name.split('/')[0] == 'module'  # type: ignore
            enabled = section[module_name].get('enabled', True)
            enabled = enabled not in (
                'false',
                'off',
                'disabled',
                'not',
                '0',
                False,
            )

            return valid_module_name and enabled
        except KeyError:
            return False

class ModuleManager:
    """A manager for operating on a set of modules."""

    def __init__(self, config: ApplicationConfig) -> None:
        self.application_config = config
        self.application_context = context(config)
        self.modules: Dict[str, Module] = {}
        self.startup_done = False
        self.last_module_periods: Dict[str, str] = {}

        # Create a dictionary containing all managed templates, mapping to
        # the tuple (module, template_shortname)
        self.managed_templates: Dict[Path, Tuple[Module, str]] = {}

        for section, options in config.items():
            module_config = {section: options}
            if Module.valid_class_section(module_config):
                module = Module(
                    module_config=module_config,
                    config_directory=self.application_config['_runtime']['config_directory'],
                    temp_directory=self.application_config['_runtime']['temp_directory'],
                )
                self.modules[module.name] = module

                # Insert the modules templates into the template Path map
                for shortname, template in module.templates.items():
                    self.managed_templates[template['source']] = (module, shortname)


        # Initialize the config directory watcher, but don't start it yet
        self.directory_watcher = DirectoryWatcher(
            directory=self.application_config['_runtime']['config_directory'],
            on_modified=self.modified,
        )

    def __len__(self) -> int:
        """Return the number of managed modules."""
        return len(self.modules)

    def module_periods(self) -> Dict[str, str]:
        """Return dict containing the period of all modules."""
        module_periods = {}
        for module_name, module in self.modules.items():
            module_periods[module_name] = module.timer.period()

        return module_periods

    def finish_tasks(self) -> None:
        """
        Finish all due tasks defined by the managed modules.

        The order of finishing tasks is as follows:
            1) Import any relevant context sections.
            2) Compile all templates with the new section.
            3) Run startup commands, if it is not already done.
            4) Run period change commands, if it is not already done for this
               module periods combination.
        """
        if not self.startup_done:
            self.import_context_sections('on_startup')
            self.compile_templates('on_startup')
            self.startup()
            self.period_change()
        elif self.last_module_periods != self.module_periods():
            self.import_context_sections('on_period_change')
            self.compile_templates('on_period_change')
            self.period_change()

    def has_unfinished_tasks(self) -> bool:
        """Return True if there are any module tasks due."""
        if not self.startup_done:
            return True
        else:
            return self.last_module_periods != self.module_periods()

    def time_until_next_period(self) -> timedelta:
        """Time left until first period change of any of the modules managed."""
        return min(
            module.timer.time_until_next_period()
            for module
            in self.modules.values()
        )

    def import_context_sections(self, trigger: str) -> None:
        """
        Import context sections defined by the managed modules.

        Trigger is one of 'on_startup', 'on_period_change', or 'on_exit'.
        This determines which section of the module is used to get the context
        import specification from.
        """
        assert trigger in ('on_startup', 'on_period_change', 'on_exit',)

        for module in self.modules.values():
            context_section_imports = module.context_section_imports(trigger)

            for csi in context_section_imports:
                self.application_context = insert_into(
                    context=self.application_context,
                    section=csi.into_section,
                    from_section=csi.from_section,
                    from_config_file=csi.from_config_file,
                )

    def compile_templates(self, trigger: str) -> None:
        """
        Compile the module templates specified by the `templates` option.

        Trigger is one of 'on_startup', 'on_period_change', or 'on_exit'.
        This determines which section of the module is used to get the compile
        specification from.
        """
        assert trigger in ('on_startup', 'on_period_change', 'on_exit',)

        for module in self.modules.values():
            for shortname in module.module_config.get(trigger, {}).get('compile', []):
                self.compile_template(module=module, shortname=shortname)


    def compile_template(self, module: Module, shortname: str) -> None:
        """
        Compile a single template given by its shortname.

        A shortname is given either by shortname, implying that the module given
        defines that template, or by module_name.shortname, making it explicit.
        """
        *_module, template_name = shortname.split('.')
        if len(_module) == 1:
            # Explicit module has been specified
            module_name = _module[0]
        else:
            # No module has been specified, use the module itself
            module_name = module.name

        compiler.compile_template(  # type: ignore
            template=self.modules[module_name].templates[template_name]['source'],
            target=self.modules[module_name].templates[template_name]['target'],
            context=self.application_context,
            shell_command_working_directory=self.application_config['_runtime']['config_directory'],
        )

    def startup(self):
        """Run all startup commands specified by the managed modules."""
        for module in self.modules.values():
            startup_commands = module.startup_commands()
            for command in startup_commands:
                logger.info(f'[module/{module.name}] Running startup command.')
                self.run_shell(command, module.name)

        self.startup_done = True

        # Start watching config directory for file changes
        self.directory_watcher.start()

    def period_change(self):
        """Run all period change commands specified by the managed modules."""
        for module in self.modules.values():
            period_change_commands = module.period_change_commands()
            for command in period_change_commands:
                logger.info(f'[module/{module.name}] Running period change command.')
                self.run_shell(command, module.name)

        self.last_module_periods = self.module_periods()

    def exit(self):
        """
        Run all exit commands specified by the managed modules.

        Also close all temporary file handlers created by the modules.
        """
        for module in self.modules.values():
            exit_commands = module.exit_commands()
            for command in exit_commands:
                logger.info(f'[module/{module.name}] Running exit command.')
                self.run_shell(command, module.name)

            if hasattr(module, 'temp_files'):
                for temp_file in module.temp_files:
                    temp_file.close()

        # Stop watching config directory for file changes
        self.directory_watcher.stop()

    def modified(self, modified: Path):
        """
        Callback for when files within the config directory are modified.

        Run any context imports, compilations, and shell commands specified
        within the on_modified event block of each module.
        """
        if not modified in self.managed_templates:
            # The modified file is not specified in any of the modules
            return

        module, template_shortname = self.managed_templates[modified]

        if template_shortname in module.module_config.get('on_modified', {}):
            for shortname in module.module_config['on_modified'][template_shortname].get('compile', []):
                self.compile_template(
                    shortname=shortname,
                    module=module,
                )

        modified_commands = module.modified_commands(template_shortname)
        for command in modified_commands:
            logger.info(f'[module/{module.name}] Running modified command.')
            self.run_shell(command, module.name)

    def run_shell(self, command: str, module_name: str) -> None:
        """Run a shell command defined by a managed module."""
        logger.info(f'[module/{module_name}] Running command "{command}".')
        run_shell(
            command=command,
            working_directory=self.application_config['_runtime']['config_directory'],
        )




"""Application wide fixtures."""
import os
from pathlib import Path

import pytest

from astrality.config import generate_expanded_env_dict, user_configuration


@pytest.fixture
def conf_path():
    """Return str path to configuration directory."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[1] / 'config'
    return conf_path


@pytest.fixture
def conf_file_path(conf_path):
    """Return path to example configuration."""
    return conf_path / 'astrality.yaml'


@pytest.fixture(scope='session', autouse=True)
def conf():
    """Return the configuration object for the example configuration."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[2]
    return user_configuration(conf_path)


@pytest.fixture
def expanded_env_dict():
    """Return expanded environment dictionary."""
    return generate_expanded_env_dict()



"""Module for compilation of templates."""

from functools import partial
import logging
from pathlib import Path
from typing import Any, Dict, Union

from jinja2 import (
    Environment,
    FileSystemLoader,
    Undefined,
    make_logging_undefined,
)

from astrality.resolver import Resolver
from astrality.utils import run_shell


Context = Dict[str, Resolver]
ApplicationConfig = Dict[str, Dict[str, Any]]

logger = logging.getLogger('astrality')


def cast_to_numeric(value: str) -> Union[int, float, str]:
    """Casts string to numeric type if possible, else return string."""
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


def context(config: ApplicationConfig) -> Context:
    """
    Return a context dictionary based on the contents of a config dict.

    Only sections named context/* are considered to be context sections, and
    these sections are returned with 'context/' stripped away, and with their
    contents cast to a Resolver instance.
    """
    contents: Context = {}
    for section_name, section in config.items():
        if not isinstance(section_name, str) or \
           not len(section_name) > 8 or \
           not section_name[:8] == 'context/':
            continue
        else:
            category = section_name[8:]
            contents[category] = Resolver(section)

    return contents


def jinja_environment(
    templates_folder: Path,
    shell_command_working_directory: Path,
) -> Environment:
    """Return a jinja Environment instance for templates in a folder."""
    logger = logging.getLogger(__name__)
    LoggingUndefined = make_logging_undefined(
        logger=logger,
        base=Undefined
    )

    env = Environment(
        loader=FileSystemLoader(
            str(templates_folder),
            followlinks=True,
        ),
        autoescape=False,
        auto_reload=True,
        optimized=True,
        finalize=finalize_variable_expression,
        undefined=LoggingUndefined,
    )

    # Add run shell command filter
    run_shell_from_working_directory = partial(
        run_shell,
        working_directory=shell_command_working_directory,
    )
    env.filters['shell'] = run_shell_from_working_directory

    return env


def finalize_variable_expression(result: str) -> str:
    """Return empty strings for undefined template variables."""
    if result is None:
        return ''
    else:
        return result


def compile_template(
    template: Path,
    target: Path,
    context: Context,
    shell_command_working_directory: Path,
) -> None:
    """Compile template to target destination with specific context."""
    logger.info(f'[Compiling] Template: "{template}" -> Target: "{target}"')

    env = jinja_environment(
        templates_folder=template.parent,
        shell_command_working_directory=shell_command_working_directory,
    )
    jinja_template = env.get_template(name=template.name)
    result = jinja_template.render(context)

    with open(target, 'w') as target_file:
        target_file.write(result)



from astrality import compiler

import pytest


@pytest.fixture
def context_config():
    return {
        'module/whatever': {},
        'nonsenical_entry': 5,
        'context/general': {
            'font': 'comic sans',
            'monitor': 'HDMI1',
        },
        'context/colors': {
            'foreground': 'white',
            'background': 'black',
            1: 'white',
            2: 'black',
        },
    }


@pytest.fixture
def context(context_config):
    return compiler.context(context_config)


def test_context_contents(context):
    assert context == {
        'general': {
            'font': 'comic sans',
            'monitor': 'HDMI1',
        },
        'colors': {
            'foreground': 'white',
            'background': 'black',
            1: 'white',
            2: 'black',
        },
    }


def test_context_integer_index_resolution(context):
    assert context['colors'][3] == 'black'


def test_context_initialization_with_no_context_blocks():
    application_config = {'module/whatever': {'timer': {'type': 'solar'}}}
    assert compiler.context(application_config) == {}



import logging
import os
import time
from datetime import datetime, timedelta
from pathlib import Path

from freezegun import freeze_time
import pytest

from astrality import timer
from astrality.config import generate_expanded_env_dict
from astrality.module import ContextSectionImport, Module, ModuleManager
from astrality.resolver import Resolver


@pytest.fixture
def valid_module_section():
    return {
        'module/test_module': {
            'enabled': True,
            'timer': {'type': 'weekday'},
            'templates': {
                'template_name': {
                    'source': '../tests/templates/test_template.conf',
                    'target': '/tmp/compiled_result',
                }
            },
            'on_startup': {
                'run': ['echo {period}'],
                'compile': ['template_name'],
            },
            'on_period_change': {'run': ['echo {template_name}']},
            'on_exit': {'run': ['echo exit']},
        }
    }


@pytest.fixture
def folders(conf):
    return (
        conf['_runtime']['config_directory'],
        conf['_runtime']['temp_directory'],
    )

@pytest.fixture
def simple_application_config(valid_module_section, folders, expanded_env_dict):
    config = valid_module_section.copy()
    config['_runtime'] = {}
    config['_runtime']['config_directory'], \
        config['_runtime']['temp_directory'] = folders
    config['context/env'] = expanded_env_dict
    config['context/fonts'] = {1: 'FuraMono Nerd Font'}
    return config


@pytest.fixture
def module(valid_module_section, folders):
    return Module(valid_module_section, *folders)

@pytest.fixture
def single_module_manager(simple_application_config):
    return ModuleManager(simple_application_config)


class TestModuleClass:

    def test_valid_class_section_method_with_valid_section(self, valid_module_section):
        assert Module.valid_class_section(valid_module_section) == True

    def test_valid_class_section_method_with_disabled_module_section(self):
        disabled_module_section =  {
            'module/disabled_test_module': {
                'enabled': False,
                'on_startup': {'run': ['test']},
                'on_period_change': {'run': ['']},
                'on_exit': {'run': ['whatever']},
            }
        }
        assert Module.valid_class_section(disabled_module_section) == False

    def test_valid_class_section_method_with_invalid_section(self):
        invalid_module_section =  {
            'context/fonts': {
                'some_key': 'some_value',
            }
        }
        assert Module.valid_class_section(invalid_module_section) == False

    def test_valid_class_section_with_wrongly_sized_dict(self, valid_module_section):
        invalid_module_section = valid_module_section
        invalid_module_section.update({'module/valid2': {'enabled': True}})

        with pytest.raises(RuntimeError):
            Module.valid_class_section(invalid_module_section)

    def test_module_name(self, module):
        assert module.name == 'test_module'

    def test_module_timer_class(self, module):
        assert isinstance(module.timer, timer.Weekday)

    def test_using_default_static_timer_when_no_timer_is_given(self, folders):
        static_module = Module({'module/static': {}}, *folders)
        assert isinstance(static_module.timer, timer.Static)

    @freeze_time('2018-01-27')
    def test_get_shell_commands_with_special_interpolations(
        self,
        module,
        caplog,
    ):
        assert module.startup_commands() == ('echo saturday',)

        compilation_target = '/tmp/compiled_result'
        assert module.period_change_commands() == (
            f'echo {compilation_target}',
        )

    @freeze_time('2018-01-27')
    def test_running_module_manager_commands_with_special_interpolations(
        self,
        single_module_manager,
        caplog,
    ):
        single_module_manager.startup()
        assert (
            'astrality',
            logging.INFO,
            '[module/test_module] Running command "echo saturday".',
        ) in caplog.record_tuples
        assert (
            'astrality',
            logging.INFO,
            'saturday\n',
        ) in caplog.record_tuples

        caplog.clear()
        single_module_manager.period_change()
        compilation_target = '/tmp/compiled_result'
        assert (
            'astrality',
            logging.INFO,
            '[module/test_module] Running command "echo /tmp/compiled_result".',
        ) in caplog.record_tuples
        assert (
            'astrality',
            logging.INFO,
            compilation_target + '\n',
        ) in caplog.record_tuples

    @pytest.mark.slow
    def test_running_shell_command_that_times_out(self, single_module_manager, caplog):
        single_module_manager.run_shell('sleep 2.1', 'name')
        assert 'used more than 2 seconds' in caplog.record_tuples[1][2]

    def test_running_shell_command_with_non_zero_exit_code(
        self,
        single_module_manager,
        caplog,
    ):
        single_module_manager.run_shell('thiscommandshould not exist', 'name')
        assert 'not found' in caplog.record_tuples[1][2]
        assert 'non-zero return code' in caplog.record_tuples[2][2]

    def test_running_shell_command_with_environment_variable(
        self,
        single_module_manager,
        caplog,
    ):
        single_module_manager.run_shell('echo $USER', 'name')
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                '[module/name] Running command "echo $USER".',
            ),
            (
                'astrality',
                logging.INFO,
                os.environ['USER'] + '\n',
            )
        ]

    @freeze_time('2018-01-27')
    def test_running_module_startup_command(
        self,
        single_module_manager,
        module,
        valid_module_section,
        caplog,
    ):
        single_module_manager.startup()

        template_file = str(module.templates['template_name']['source'])
        compiled_template = str(module.templates['template_name']['target'])

        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running startup command.',
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running command "echo saturday".',
            ),
            (
                'astrality',
                logging.INFO,
                'saturday\n',
            )
        ]

    def test_running_module_startup_command_when_no_command_is_specified(
        self,
        simple_application_config,
        module,
        caplog,
    ):
        simple_application_config['module/test_module']['on_startup'].pop('run')
        module_manager = ModuleManager(simple_application_config)

        module_manager.startup()

        template_file = str(module.templates['template_name']['source'])
        compiled_template = str(module.templates['template_name']['target'])

        assert caplog.record_tuples == [
            (
                'astrality',
                logging.DEBUG,
                '[module/test_module] No startup command specified.',
            ),
        ]

    def test_running_module_period_change_command(
        self,
        single_module_manager,
        module,
        caplog,
    ):
        single_module_manager.period_change()

        template_file = str(module.templates['template_name']['source'])
        compiled_template = str(module.templates['template_name']['target'])

        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running period change command.',
            ),
            (
                'astrality',
                logging.INFO,
                f'[module/test_module] Running command "echo {compiled_template}".',
            ),
            (
                'astrality',
                logging.INFO,
                f'{compiled_template}\n',
            )
        ]

    def test_running_module_period_change_command_when_no_command_is_specified(
        self,
        simple_application_config,
        module,
        conf,
        caplog,
    ):
        simple_application_config['module/test_module']['on_period_change'].pop('run')
        module_manager = ModuleManager(simple_application_config)

        module_manager.period_change()

        template_file = str(module.templates['template_name']['source'])
        compiled_template = str(module.templates['template_name']['target'])

        assert caplog.record_tuples == [
            (
                'astrality',
                logging.DEBUG,
                '[module/test_module] No period change command specified.',
            ),
        ]

    def test_running_module_exit_command(self, single_module_manager, caplog):
        single_module_manager.exit()
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running exit command.',
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running command "echo exit".',
            ),
            (
                'astrality',
                logging.INFO,
                'exit\n',
            )
        ]

    def test_running_module_exit_command_when_no_command_is_specified(
        self,
        simple_application_config,
        caplog,
    ):
        simple_application_config['module/test_module']['on_exit'].pop('run')
        module_manager = ModuleManager(simple_application_config)

        module_manager.exit()
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.DEBUG,
                '[module/test_module] No exit command specified.',
            ),
        ]

    @freeze_time('2018-02-05')
    def test_running_finished_tasks_command(
        self,
        simple_application_config,
        caplog,
    ):
        """Test that every task is finished at first finish_tasks() invocation."""
        module_manager = ModuleManager(simple_application_config)
        module_manager.finish_tasks()
        template = str(module_manager.modules['test_module'].templates['template_name']['source'])
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                f'[Compiling] Template: "{template}" -> Target: "/tmp/compiled_result"'
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running startup command.',
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running command "echo monday".',
            ),
            (
                'astrality',
                logging.INFO,
                'monday\n',
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running period change command.',
            ),
            (
                'astrality',
                logging.INFO,
                '[module/test_module] Running command "echo /tmp/compiled_result".',
            ),
            (
                'astrality',
                logging.INFO,
                '/tmp/compiled_result\n',
            )
        ]

    def test_location_of_template_file_defined_relatively(self, module):
        template_file = module.templates['template_name']['source']
        compiled_template = module.templates['template_name']['target']

        assert template_file.resolve() == Path(__file__).parent / 'templates' / 'test_template.conf'
        assert compiled_template == Path('/tmp/compiled_result')

    def test_location_of_template_file_defined_absolutely(
        self,
        valid_module_section,
        folders,
    ):
        absolute_path = Path(__file__).parent / 'templates' / 'test_template.conf'
        valid_module_section['module/test_module']['templates']['template_name']['source'] = absolute_path

        module = Module(valid_module_section, *folders)
        template_file = module.templates['template_name']['source']

        assert template_file == absolute_path

    def test_missing_template_file(
        self,
        valid_module_section,
        folders,
        caplog,
    ):
        valid_module_section['module/test_module']['templates']['template_name']['source'] = \
            '/not/existing'

        module = Module(valid_module_section, *folders)
        assert module.templates == {}
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.ERROR,
                '[module/test_module] Template "template_name": source "/not/existing" does'
                ' not exist. Skipping compilation of this file.'
            ),
        ]

    def test_expand_path_method(self, module, conf):
        absolute_path = Path('/tmp/ast')
        tilde_path = Path('~/dir')
        relative_path = Path('test')
        assert module.expand_path(absolute_path) == absolute_path
        assert module.expand_path(tilde_path) == Path.home() / 'dir'
        assert module.expand_path(relative_path) == \
            conf['_runtime']['config_directory'] / 'test'

    def test_create_temp_file_method(self, module):
        temp_file = module.create_temp_file()
        assert temp_file.is_file()

    def test_cleanup_of_tempfile_on_exit(self, single_module_manager):
        temp_file = single_module_manager.modules['test_module'].create_temp_file()
        assert temp_file.is_file()
        single_module_manager.exit()
        assert not temp_file.is_file()

    def test_creation_of_temporary_file_when_compiled_template_is_not_defined(
        self,
        simple_application_config,
    ):
        simple_application_config['module/test_module']['templates']['template_name'].pop('target')
        module_manager = ModuleManager(simple_application_config)
        assert module_manager.modules['test_module'].templates['template_name']['target'].is_file()

    def test_compilation_of_template(
        self,
        simple_application_config,
        module,
        conf,
        caplog,
    ):
        simple_application_config['module/test_module']['timer']['type'] = 'solar'
        compiled_template_content = 'some text\n' + os.environ['USER'] + '\nFuraMono Nerd Font'
        module_manager = ModuleManager(simple_application_config)
        module_manager.compile_templates('on_startup')

        template_file = str(module.templates['template_name']['source'])
        compiled_template = str(module.templates['template_name']['target'])

        with open('/tmp/compiled_result', 'r') as file:
            compiled_result = file.read()

        assert compiled_template_content == compiled_result
        assert caplog.record_tuples == [
            (
                'astrality',
                logging.INFO,
                f'[Compiling] Template: "{template_file}" -> Target: "{compiled_template}"'
            ),
        ]

def test_has_unfinished_tasks(simple_application_config, freezer):
    # Move time to midday
    midday = datetime.now().replace(hour=12, minute=0)
    freezer.move_to(midday)

    # At instanziation, the module should have unfinished tasks
    weekday_module = ModuleManager(simple_application_config)
    assert weekday_module.has_unfinished_tasks() == True

    # After finishing tasks, there should be no unfinished tasks (duh!)
    weekday_module.finish_tasks()
    assert weekday_module.has_unfinished_tasks() == False

    # If we move the time forwards, but not to a new period, there should still
    # not be any unfinished tasks
    before_midnight = datetime.now().replace(hour=23, minute=59)
    freezer.move_to(before_midnight)
    assert weekday_module.has_unfinished_tasks() == False

    # But right after a period change (new weekday), there should be unfinished
    # tasks
    two_minutes = timedelta(minutes=2)
    freezer.move_to(before_midnight + two_minutes)
    assert weekday_module.has_unfinished_tasks() == True

    # Again, after finishing tasks, there should be no unfinished tasks left
    weekday_module.finish_tasks()
    assert weekday_module.has_unfinished_tasks() == False



@pytest.fixture
def config_with_modules():
    return {
        'context/env': generate_expanded_env_dict(),
        'module/solar_module': {
            'enabled': True,
            'timer': {
                'type': 'solar',
                'longitude': 0,
                'latitude': 0,
                'elevation': 0,
            },
            'templates': {
                'template_name': {
                    'source': 'astrality/tests/templates/test_template.conf',
                    'target': '/tmp/compiled_result',
                }
            },
            'on_startup': {'run': ['echo solar compiling {template_name}']},
            'on_period_change': {'run': ['echo solar {period}']},
            'on_exit': {'run': ['echo solar exit']},
        },
        'module/weekday_module': {
            'enabled': True,
            'timer': {'type': 'weekday'},
            'on_startup': {'run': ['echo weekday startup']},
            'on_period_change': {'run': ['echo weekday {period}']},
            'on_exit': {'run': ['echo weekday exit']},
        },
        'module/disabled_module': {
            'enabled': False,
            'timer': 'static',
        },
        'context/fonts': {1: 'FuraCode Nerd Font'},
        '_runtime': {
            'config_directory': Path(__file__).parents[2],
            'temp_directory': '/tmp',
        }
    }

@pytest.fixture
def module_manager(config_with_modules):
    return ModuleManager(config_with_modules)


def test_import_sections_on_period_change(config_with_modules, freezer):
    config_with_modules['module/weekday_module']['on_period_change']['import_context'] = [{
        'to_section': 'week',
        'from_file': 'astrality/tests/templates/weekday.yaml',
        'from_section': '{period}',
    }]
    config_with_modules.pop('module/solar_module')
    module_manager = ModuleManager(config_with_modules)

    assert module_manager.application_context['fonts'] == {1: 'FuraCode Nerd Font'}

    sunday = datetime(year=2018, month=2, day=4)
    freezer.move_to(sunday)
    module_manager.finish_tasks()

    # Make application_context comparisons easier
    module_manager.application_context.pop('env')

    # Startup does not count as a period change, so no context has been imported
    assert module_manager.application_context == {
        'fonts': Resolver({1: 'FuraCode Nerd Font'}),
    }

    monday = datetime(year=2018, month=2, day=5)
    freezer.move_to(monday)
    module_manager.finish_tasks()

    # The period has now changed, so context should be imported
    assert module_manager.application_context == {
        'fonts': Resolver({1: 'FuraCode Nerd Font'}),
        'week': Resolver({'day': 'monday'}),
    }

def test_compiling_templates_on_cross_of_module_boundries():
    module_A = {
        'templates': {
            'template_A': {
                'source': '../tests/templates/no_context.template',
            },
        },
    }
    modules_config = {
        'module/A': module_A,
        '_runtime': {
            'config_directory': Path(__file__).parent,
            'temp_directory': Path('/tmp'),
        },
    }

    module_manager = ModuleManager(modules_config)
    module_manager.finish_tasks()

    # Modules should not compile their templates unless they explicitly
    # define a compile string in a on_* block.
    with open(module_manager.modules['A'].templates['template_A']['target']) as compilation:
        assert compilation.read() == ''

    # We now insert another module, B, which compiles the template of the
    # previous module, A
    module_B = {
        'on_startup': {
            'compile': ['A.template_A'],
        },
    }
    modules_config['module/B'] = module_B
    module_manager = ModuleManager(modules_config)
    module_manager.finish_tasks()
    with open(module_manager.modules['A'].templates['template_A']['target']) as compilation:
        assert compilation.read() == 'one\ntwo\nthree'


def test_import_sections_on_startup(config_with_modules, freezer):
    # Insert day the module was started into 'start day'
    config_with_modules['module/weekday_module']['on_startup']['import_context'] = [{
        'to_section': 'start_day',
        'from_file': 'astrality/tests/templates/weekday.yaml',
        'from_section': '{period}',
    }]

    # Insert the current day into 'day_now'
    config_with_modules['module/weekday_module']['on_period_change']['import_context'] = [{
        'to_section': 'day_now',
        'from_file': 'astrality/tests/templates/weekday.yaml',
        'from_section': '{period}',
    }]
    config_with_modules.pop('module/solar_module')
    module_manager = ModuleManager(config_with_modules)

    # Remove 'env' context for easier comparisons
    module_manager.application_context.pop('env')

    # Before finishing tasks, no context sections are imported
    assert module_manager.application_context['fonts'] == {1: 'FuraCode Nerd Font'}

    # Start module on a monday
    sunday = datetime(year=2018, month=2, day=4)
    freezer.move_to(sunday)
    module_manager.finish_tasks()
    assert module_manager.application_context == {
        'fonts': Resolver({1: 'FuraCode Nerd Font'}),
        'start_day': Resolver({'day': 'sunday'}),
    }

    # 'now_day' should now be added, but 'start_day' should remain unchanged
    monday = datetime(year=2018, month=2, day=5)
    freezer.move_to(monday)
    module_manager.finish_tasks()
    assert module_manager.application_context == {
        'fonts': Resolver({1: 'FuraCode Nerd Font'}),
        'start_day': Resolver({'day': 'sunday'}),
        'day_now': Resolver({'day': 'monday'}),
    }


def test_context_section_imports(folders):
    module_config = {
        'module/name': {
            'on_startup': {
                'import_context': [
                    {
                        'from_file': '/testfile',
                        'from_section': 'source_section',
                        'to_section': 'target_section',
                    }
                ]
            },
            'on_period_change': {
                'import_context': [
                    {
                        'from_file': '/testfile',
                        'from_section': 'source_section',
                    }
                ]
            },
        },
    }
    module = Module(module_config, *folders)
    startup_csis = module.context_section_imports('on_startup')
    expected = (
        ContextSectionImport(
            from_config_file=Path('/testfile'),
            from_section='source_section',
            into_section='target_section',
        ),
    )
    assert startup_csis == expected

    period_change_csis = module.context_section_imports('on_period_change')
    expected = (
        ContextSectionImport(
            from_config_file=Path('/testfile'),
            from_section='source_section',
            into_section='source_section',
        ),
    )
    assert period_change_csis == expected


class TestModuleManager:
    def test_invocation_of_module_manager_with_config(self, conf):
        ModuleManager(conf)

    @pytest.mark.slow
    def test_using_finish_tasks_on_example_configuration(self, conf):
        module_manager = ModuleManager(conf)
        module_manager.finish_tasks()

    def test_number_of_modules_instanziated_by_module_manager(self, module_manager):
        assert len(module_manager) == 2


def test_time_until_next_period_of_several_modules(config_with_modules, module_manager, freezer):
    solar_timer = timer.Solar(config_with_modules)
    noon = solar_timer.location.sun()['noon']

    one_minute = timedelta(minutes=1)
    freezer.move_to(noon - one_minute)

    assert module_manager.time_until_next_period() == one_minute
    two_minutes_before_midnight = datetime.now().replace(hour=23, minute=58)
    freezer.move_to(two_minutes_before_midnight)

    assert module_manager.time_until_next_period().total_seconds() == \
                              timedelta(minutes=2).total_seconds()

def test_detection_of_new_period_involving_several_modules(
    config_with_modules,
    freezer,
):
    # Move time to right before noon
    solar_timer = timer.Solar(config_with_modules)
    noon = solar_timer.location.sun()['noon']
    one_minute = timedelta(minutes=1)
    freezer.move_to(noon - one_minute)
    module_manager = ModuleManager(config_with_modules)

    # All modules should now considered period changed
    assert module_manager.has_unfinished_tasks() == True

    # Running period change method for all the period changed modules
    module_manager.finish_tasks()

    # After running these methods, they should all be reverted to not changed
    assert module_manager.has_unfinished_tasks() == False

    # Move time to right after noon
    freezer.move_to(noon + one_minute)

    # The solar timer should now be considered to have been period changed
    assert module_manager.has_unfinished_tasks() == True

    # Again, check if period_change() method makes them unchanged
    module_manager.finish_tasks()
    assert module_manager.has_unfinished_tasks() == False

    # Move time two days forwards
    two_days = timedelta(days=2)
    freezer.move_to(noon + two_days)

    # Now both timers should be considered period changed
    assert module_manager.has_unfinished_tasks() == True

def test_that_shell_filter_is_run_from_config_directory(conf_path):
    shell_filter_template = Path(__file__).parent / 'templates' / 'shell_filter_working_directory.template'
    shell_filter_template_target = Path('/tmp/astrality/shell_filter_working_directory.template')
    config = {
        'module/A': {
            'templates': {
                'shell_filter_template': {
                    'source': str(shell_filter_template),
                    'target': str(shell_filter_template_target),
                },
            },
            'on_startup': {
                'compile': ['shell_filter_template'],
            },
        },
        '_runtime': {
            'config_directory': conf_path,
            'temp_directory': Path('/tmp/astrality'),
        },
    }
    module_manager = ModuleManager(config)
    module_manager.compile_templates('on_startup')

    with open(shell_filter_template_target) as compiled:
        assert compiled.read() == str(conf_path)

    os.remove(shell_filter_template_target)


@pytest.yield_fixture
def modules_config(conf_path):
    empty_template = Path(__file__).parent / 'templates' / 'empty.template'
    empty_template_target = Path('/tmp/astrality/empty_temp_template')
    temp_directory = Path('/tmp/astrality')
    touch_target = temp_directory / 'touched'

    secondary_template = Path(__file__).parent / 'templates' / 'no_context.template'
    secondary_template_target = temp_directory / 'secondary_template.tmp'

    config = {
        'module/A': {
            'templates': {
                'template1': {
                    'source': str(empty_template),
                    'target': str(empty_template_target),
                },
            },
            'on_modified': {
                'template1': {
                    'compile': ['template1', 'B.template1'],
                    'run': ['touch ' + str(touch_target)],
                },
            },
        },
        'module/B': {
            'templates': {
                'template1': {
                    'source': str(secondary_template),
                    'target': str(secondary_template_target),
                },
            },
        },
        '_runtime': {
            'config_directory': Path(__file__).parent,
            'temp_directory': temp_directory,
        }
    }
    yield (
        config,
        empty_template,
        empty_template_target,
        touch_target,
        secondary_template,
        secondary_template_target,
    )

    # Cleanup all test files for next test iteration
    if empty_template.is_file():
        empty_template.write_text('')
    if empty_template_target.is_file():
        os.remove(empty_template_target)
    if secondary_template_target.is_file():
        os.remove(secondary_template_target)
    if touch_target.is_file():
        os.remove(touch_target)

class TestModuleFileWatching:
    def test_modified_commands_of_module(self, modules_config):
        config, empty_template, empty_template_target, touch_target, *_= modules_config
        module_manager = ModuleManager(config)
        assert module_manager.modules['A'].modified_commands('template1') == \
            ('touch ' + str(touch_target), )

    def test_direct_invocation_of_modifed_method_of_module_manager(self, modules_config):
        (
            config,
            empty_template,
            empty_template_target,
            touch_target,
            secondary_template,
            secondary_template_target,
        ) = modules_config
        module_manager = ModuleManager(config)

        # PS: Disabling the directory watcher is not necessary, as it is done in
        # the startup method.

        # Now write new text to the template
        empty_template.write_text('new content')

        # And trigger the modified method manually
        module_manager.modified(empty_template)

        # And assert that the new template has been compiled
        assert empty_template_target.is_file()
        with open(empty_template_target) as file:
            assert file.read() == 'new content'

        # And that the new file has been touched
        assert touch_target.is_file()

    def test_on_modified_event_in_module(self, modules_config):
        (
            config,
            empty_template,
            empty_template_target,
            touch_target,
            secondary_template,
            secondary_template_target,
        ) = modules_config
        module_manager = ModuleManager(config)

        # Start the file watcher by invoking the startup command indirectly
        # through finish_tasks() method
        module_manager.finish_tasks()

        # Assert that the template file is really empty as a sanity check
        with open(empty_template) as file:
            assert file.read() == ''

        # And that target files are not created yet
        assert not touch_target.is_file()
        assert not empty_template_target.is_file()
        assert not secondary_template_target.is_file()

        # Trigger the on_modified event
        empty_template.write_text('new content')
        time.sleep(1)

        # And assert that the new template has been compiled
        assert empty_template_target.is_file()
        with open(empty_template_target) as file:
            assert file.read() == 'new content'

        # Assert that also templates from other modules are compiled
        assert secondary_template_target.is_file()
        with open(secondary_template_target) as file:
            assert file.read() == 'one\ntwo\nthree'

        # And that the new file has been touched
        assert touch_target.is_file()



import os
import shutil
import time
from pathlib import Path

import pytest

from astrality.filewatcher import DirectoryWatcher


@pytest.yield_fixture
def test_files():
    """Return paths related to two test files and cleanup afterwards."""
    watched_directory = Path('/tmp/astrality')
    test_file1 = watched_directory / 'tmp_test_file1'

    recursive_dir = watched_directory / 'test_folder'
    test_file2 = recursive_dir / 'tmp_test_file2'

    yield watched_directory, recursive_dir, test_file1, test_file2

    # Cleanup files
    if test_file1.is_file():
        os.remove(test_file1)
    if test_file2.is_file():
        os.remove(test_file2)
    if recursive_dir.is_dir():
        shutil.rmtree(recursive_dir)


@pytest.yield_fixture
def watch_dir():
    """Instanciate a directory watcher and stop it after its use."""
    class EventSaver:
        """Mock class for testing callback function."""

        def __init__(self):
            self.called = 0

        def save_argument(self, path: Path) -> None:
            self.called += 1
            self.argument = path

    event_saver = EventSaver()

    # Watch a temporary directory
    watched_directory = Path('/tmp/astrality')
    dir_watcher = DirectoryWatcher(
        directory=watched_directory,
        on_modified=event_saver.save_argument,
    )

    yield dir_watcher, event_saver

    dir_watcher.stop()


@pytest.mark.slow
def test_filesystem_watcher(test_files, watch_dir):
    """Test correct callback invocation on directory watching."""
    watched_directory, recursive_dir, test_file1, test_file2 = test_files
    dir_watcher, event_saver = watch_dir

    # Start watching the directory
    dir_watcher.start()

    # Nothing has been modified yet
    assert not hasattr(event_saver, 'argument')
    assert event_saver.called == 0

    # Create an empty file
    test_file1.touch()

    # New files are not considered "modified"
    time.sleep(0.7)
    assert not hasattr(event_saver, 'argument')
    assert event_saver.called == 0

    # But when we write to it, it is considered "modified"
    test_file1.write_text('test_content')
    time.sleep(0.7)
    assert event_saver.argument == test_file1
    assert event_saver.called == 1

    # Create a directory in the watched directory
    recursive_dir.mkdir(parents=True)

    # Subdirectories are not of interest
    time.sleep(0.7)
    assert event_saver.argument == test_file1
    assert event_saver.called == 1

    # Create a file in the subdirectory
    test_file2.write_text('test')

    # Both the touch event and the write event are considered of interest
    time.sleep(0.7)
    assert event_saver.argument == test_file2
    assert event_saver.called == 2



"""Tests for the compiler module."""

import logging
from pathlib import Path

import pytest
from jinja2 import Environment

from astrality.compiler import cast_to_numeric, compile_template, jinja_environment
from astrality.resolver import Resolver


@pytest.fixture
def test_templates_folder():
    return Path(__file__).parent / 'templates'


@pytest.fixture
def jinja_test_env(test_templates_folder):
    return jinja_environment(
        test_templates_folder,
        shell_command_working_directory=Path('~').resolve(),
    )


def test_rendering_environment_variables(jinja_test_env, expanded_env_dict):
    template = jinja_test_env.get_template('env_vars')
    assert template.render(env=expanded_env_dict) == \
        'test_value\nfallback_value\n'


def test_logging_undefined_variables(jinja_test_env, expanded_env_dict, caplog):
    template = jinja_test_env.get_template('env_vars')
    template.render(env=expanded_env_dict)
    assert (
        'astrality.compiler',
        logging.WARNING,
        'Template variable warning: env_UNDEFINED_VARIABLE is undefined',
    ) in caplog.record_tuples


def test_integer_indexed_templates(jinja_test_env):
    template = jinja_test_env.get_template('integer_indexed')
    context = Resolver({'section': {1: 'one', 2: 'two'}})
    assert template.render(context) == 'one\ntwo\ntwo'


def test_compilation_of_jinja_template(test_templates_folder, expanded_env_dict):
    template = test_templates_folder / 'env_vars'
    target = Path('/tmp/astrality') / template.name
    context = {'env': expanded_env_dict}
    compile_template(template, target, context, Path('/'))

    with open(target) as target:
        assert target.read() == 'test_value\nfallback_value\n'

@pytest.mark.parametrize(('string,cast,resulting_type'), [
    ('-2', -2, int),
    ('0', 0, int),
    ('1', 1, int),
    ('1.5', 1.5, float),
    ('one', 'one', str),
])
def test_cast_to_numeric(string, cast, resulting_type):
    result = cast_to_numeric(string)
    assert result == cast
    assert isinstance(result, resulting_type)

def test_run_shell_template_filter(test_templates_folder):
    shell_template_path = test_templates_folder / 'shell_filter.template'
    compiled_shell_template_path = Path('/tmp/astrality') / shell_template_path.name
    context = {}
    compile_template(
        template=shell_template_path,
        target=compiled_shell_template_path,
        context=context,
        shell_command_working_directory=Path('/'),
    )

    with open(compiled_shell_template_path) as target:
        assert target.read() == 'quick\nanother_quick\nslow_but_allowed\n\nfallback'

def test_working_directory_of_shell_command_filter(test_templates_folder):
    shell_template_path = test_templates_folder / 'shell_filter_working_directory.template'
    compiled_shell_template_path = Path('/tmp/astrality') / shell_template_path.name
    context = {}
    compile_template(
        template=shell_template_path,
        target=compiled_shell_template_path,
        context=context,
        shell_command_working_directory=Path('/'),
    )

    with open(compiled_shell_template_path) as target:
        assert target.read() == '/'



from datetime import timedelta

from astrality.timer import Static


def test_static_periods():
    static = Static({})
    assert 'static' in static.periods


def test_static_time_until_next_period():
    default_static = Static({})
    assert default_static.time_until_next_period() == timedelta(days=36500)



from datetime import datetime, timedelta

import pytest

from astrality.timer import Periodic


@pytest.fixture
def periodic():
    return Periodic({'type': 'periodic'})


@pytest.mark.parametrize("period, is_period", [
    ('0', True),
    ('1', True),
    ('80', True),
    ('-1', False),
    ('1.2', False),
    ('night', False),
])
def test_periodic_periods(period, is_period, periodic):
    assert (period in periodic.periods) == is_period


def test_periodic_beginning_period(freezer):
    periodic = Periodic({})
    assert periodic.period() == '0'


def test_periodic_standard_timedelta(freezer):
    default_periodic = Periodic({})
    assert default_periodic.time_until_next_period().total_seconds() == 3600


def test_using_custom_periodic_timer(freezer):
    custom_periodic_config = {
        'type': 'periodic',
        'seconds': 1,
        'minutes': 2,
        'hours': 3,
        'days': 4,
    }
    custom_periodic = Periodic(custom_periodic_config)
    assert custom_periodic.time_until_next_period() == timedelta(
        seconds=1,
        minutes=2,
        hours=3,
        days=4,
    )


def test_weekday_time_until_next_period_of_periodic_timer(freezer):
    now = datetime.now()
    freezer.move_to(now)
    default_periodic = Periodic({})
    assert default_periodic.time_until_next_period().total_seconds() == 60*60

    fourty_minutes = timedelta(minutes=40)
    freezer.move_to(now + fourty_minutes)
    assert default_periodic.time_until_next_period().total_seconds() == 20*60

    freezer.move_to(now + 2 * fourty_minutes)
    assert default_periodic.time_until_next_period().total_seconds() == 40*60


def test_enumeration_of_periodic_timer_periods(freezer):
    now = datetime.now()
    freezer.move_to(now)
    default_periodic = Periodic({})
    assert default_periodic.period() == '0'

    fourty_minutes = timedelta(minutes=40)
    freezer.move_to(now + fourty_minutes)
    assert default_periodic.period() == '0'

    freezer.move_to(now + 2 * fourty_minutes)
    assert default_periodic.period() == '1'



"""Module for directory modification watching."""

from pathlib import Path
from typing import Callable

from watchdog.events import FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer


class DirectoryWatcher:
    """A directory watcher class."""

    def __init__(
        self,
        directory: Path,
        on_modified: Callable[[Path], None],
    ) -> None:
        """
        Initialize a watcher which observes modifications in `directory`.

        on_modified: A callable which is invoked with the path of modified
                     files within `directory`.
        """
        self.on_modified = on_modified
        self.watched_directory = str(directory)
        self.observer = Observer()

    def start(self) -> None:
        """Start watching the specified directory for file modifications."""
        event_handler = DirectoryEventHandler(self.on_modified)
        self.observer.schedule(
            event_handler,
            self.watched_directory,
            recursive=True,
        )
        self.observer.start()

    def stop(self) -> None:
        """Stop watching the directory."""
        if self.observer.is_alive():
            self.observer.stop()
            self.observer.join()


class DirectoryEventHandler(FileSystemEventHandler):
    """An event handler for filesystem changes within a directory."""

    def __init__(self, on_modified: Callable[[Path], None]) -> None:
        """Initialize event handler with callback functions."""
        self._on_modified = on_modified

    def on_modified(self, event: FileModifiedEvent) -> None:
        """Call on_modified callback function on modifed event in dir."""
        if event.is_directory:
            return

        self._on_modified(Path(event.src_path).absolute())



import os
from datetime import datetime
from pathlib import Path
from shutil import rmtree

import pytest

from astrality import compiler
from astrality.config import (
    create_config_directory,
    dict_from_config_file,
    generate_expanded_env_dict,
    insert_environment_values,
    insert_command_substitutions,
    insert_into,
    resolve_config_directory,
    preprocess_configuration_file,
)
from astrality.module import ModuleManager


@pytest.fixture
def dummy_config():
    test_conf = Path(__file__).parent / 'test.yaml'
    return dict_from_config_file(test_conf)


class TestAllConfigFeaturesFromDummyConfig:
    def test_normal_variable(self, dummy_config):
        assert dummy_config['context/section1']['var1'] == 'value1'

    def test_variable_interpolation(self, dummy_config):
        assert dummy_config['context/section1']['var2'] == 'value1/value2'
        assert dummy_config['context/section2']['var3'] == 'value1'

    def test_empty_string_variable(self, dummy_config):
        assert dummy_config['context/section2']['empty_string_var'] == ''

    def test_non_existing_variable(self, dummy_config):
        with pytest.raises(KeyError):
            assert dummy_config['context/section2']['not_existing_option'] is None

    def test_environment_variable_interpolation(self, dummy_config):
        assert dummy_config['context/section3']['env_variable'] == 'test_value, hello'


def test_config_directory_name(conf):
    assert str(conf['_runtime']['config_directory'])[-7:] == '/config'


def test_name_of_config_file(conf):
    assert '/astrality.yaml' in str(conf['_runtime']['config_file'])


@pytest.mark.slow
def test_that_colors_are_correctly_imported_based_on_wallpaper_theme(conf, freezer):
    midnight = datetime(year=2018, month=1, day=31, hour=0, minute=0)
    freezer.move_to(midnight)
    module_manager = ModuleManager(conf)
    assert 'colors' not in module_manager.application_context
    module_manager.finish_tasks()
    assert module_manager.application_context['colors'] == {1: 'CACCFD', 2: '3F72E8'}

def test_environment_variable_interpolation_by_preprocessing_conf_yaml_file():
    test_conf = Path(__file__).parent / 'test.yaml'
    result = preprocess_configuration_file(test_conf)

    expected_result = \
'''context/section1:
    var1: value1
    var2: value1/value2


context/section2:
    # Comment
    var3: value1
    empty_string_var: ''

context/section3:
    env_variable: test_value, hello

context/section4:
    1: primary_value
'''
    assert expected_result == result

@pytest.mark.slow
def test_command_substition_by_preprocessing_yaml_file():
    test_conf = Path(__file__).parent / 'commands.yaml'
    result = preprocess_configuration_file(test_conf)

    expected_result = \
'''section1:
    key1: test
    key2: test_value
    key3: test_value
    key4: 
'''
    assert expected_result == result

def test_generation_of_expanded_env_dict():
    env_dict = generate_expanded_env_dict()
    assert len(env_dict) == len(os.environ)

    for name, value in os.environ.items():
        if not '$' in value:
            assert env_dict[name] == value

def test_insert_environment_variables():
    '''Pytest sets the following environment variables
        EXAMPLE_ENV_VARIABLE=test_value
        lower_case_key=lower_case_value
        UPPER_CASE_KEY=UPPER_CASE_VALUE
        conflicting_key=value1
        CONFLICTING_KEY=value2
    '''

    config_line = 'key=value-${EXAMPLE_ENV_VARIABLE}'
    expected = 'key=value-test_value'
    assert insert_environment_values(config_line) == expected

    # Test if several variables on the same line are all interpolated
    several_env_variables = 'key=${lower_case_key}-${EXAMPLE_ENV_VARIABLE}'
    expected = 'key=lower_case_value-test_value'
    assert insert_environment_values(several_env_variables) == expected

    # Check that case sensitiveness is correctly handled
    lower_case_key = 'something ${lower_case_key}'
    expected = 'something lower_case_value'
    assert insert_environment_values(lower_case_key) == expected

    upper_case_key = 'something ${UPPER_CASE_KEY}'
    expected = 'something UPPER_CASE_VALUE'
    assert insert_environment_values(upper_case_key) == expected

    # Check if interpolation is case sensitive when "conficts" occur
    lower_case_conficting_key = '${conflicting_key}'
    expected = 'value1'
    assert insert_environment_values(lower_case_conficting_key) == expected

    upper_case_conficting_key = '${CONFLICTING_KEY}'
    expected = 'value2'
    assert insert_environment_values(upper_case_conficting_key) == expected

def test_insert_context_section():
    context = compiler.context({'context/section1': {'key_one': 'value_one'}})
    assert context['section1']['key_one'] == 'value_one'

    test_config_file = Path(__file__).parent / 'test.yaml'
    context = insert_into(
        context=context,
        section='new_section',
        from_config_file=test_config_file,
        from_section='section2'
    )

    assert context['section1']['key_one'] == 'value_one'
    assert context['new_section']['var3'] == 'value1'

    context = insert_into(
        context=context,
        section='section3',
        from_config_file=test_config_file,
        from_section='section3'
    )
    assert context['section3']['env_variable'] == 'test_value, hello'

    context = insert_into(
        context=context,
        section='section1',
        from_config_file=test_config_file,
        from_section='section1'
    )
    assert context['section1']['var2'] == 'value1/value2'
    assert 'key_one' not in context['section1']

def test_insert_command_substitutions():
    string = 'some text: $(echo result)'
    assert insert_command_substitutions(
        string,
        shell_command_working_directory=Path('~').expanduser(),
    ) == 'some text: result'


class TestResolveConfigDirectory:
    def test_setting_directory_using_application_env_variable(self, monkeypatch):
        monkeypatch.setattr(
            os,
            'environ',
            {
                'ASTRALITY_CONFIG_HOME': '/test/dir',
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/test/dir')

    def test_setting_directory_using_xdg_directory_standard(self, monkeypatch):
        monkeypatch.setattr(
            os,
            'environ',
            {
                'XDG_CONFIG_HOME': '/xdg/dir',
            },
        )
        assert resolve_config_directory() == Path('/xdg/dir/astrality')

    def test_using_standard_config_dir_when_nothing_else_is_specified(self, monkeypatch):
        monkeypatch.setattr(os, 'environ', {})
        assert resolve_config_directory() == Path('~/.config/astrality').expanduser()


class TestCreateConfigDirectory:
    def test_creation_of_empty_config_directory(self):
        config_path = Path('/tmp/config_test')
        config_dir = create_config_directory(path=config_path, empty=True)
        assert config_path == config_dir
        assert config_dir.is_dir()
        assert len(list(config_dir.iterdir())) == 0
        config_dir.rmdir()

    def test_creation_of_infered_config_directory(self, monkeypatch):
        config_path = Path('/tmp/astrality_config')
        monkeypatch.setattr(
            os,
            'environ',
            {'ASTRALITY_CONFIG_HOME': str(config_path)},
        )
        created_config_dir = create_config_directory(empty=True)
        assert created_config_dir == config_path
        created_config_dir.rmdir()

    def test_creation_of_config_directory_with_example_content(self):
        """Test copying example configuration contents."""
        config_path = Path('/tmp/astrality_config_with_contents')
        created_config_dir = create_config_directory(config_path)
        assert created_config_dir == config_path

        # Test presence of content in created folder
        dir_contents = tuple(file.name for file in created_config_dir.iterdir())
        assert 'astrality.yaml' in dir_contents
        assert 'templates' in dir_contents
        rmtree(created_config_dir)



"""Specifies everything related to application spanning configuration."""

import logging
import os
import re
from distutils.dir_util import copy_tree
from io import StringIO
from pathlib import Path
from typing import Any, Dict, Match, MutableMapping, Optional, Tuple

from astrality import compiler
from astrality.resolver import Resolver
from astrality.utils import run_shell

Context = Dict[str, Resolver]

logger = logging.getLogger('astrality')

from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper  # type: ignore
    logger.info('Using LibYAML bindings for faster .yaml parsing.')
except ImportError:  # pragma: no cover
    from yaml import Loader, Dumper
    logger.warning(
        'LibYAML not installed.'
        'Using somewhat slower pure python implementation.',
    )


ApplicationConfig = Dict[str, Dict[str, Any]]


def resolve_config_directory() -> Path:
    """
    Return the absolute configuration directory path for the application.

    The directory path is resolved as follows:
        1) If $ASTRALITY_CONFIG_HOME is present, use it.
        2) If $XDG_CONFIG_HOME is present, use $XDG_CONFIG_HOME/astrality.
        3) Elsewise, use ~/config/astrality.
    """
    if 'ASTRALITY_CONFIG_HOME' in os.environ:
        # The user has set a custom config directory for astrality
        config_directory = Path(os.environ['ASTRALITY_CONFIG_HOME'])
    else:
        # Follow the XDG directory standard
        config_directory = Path(
            os.getenv('XDG_CONFIG_HOME', '~/.config'),
            'astrality',
        )
    return config_directory.expanduser().absolute()


def infer_config_location(
    config_directory: Optional[Path] = None,
) -> Tuple[Path, Path]:
    """
    Return path of Astrality configuration file based on config folder path.

    Try to find the configuration directory and file for astrality, based on
    filesystem or specific environment variables if they are present several
    places to put it. If the expected config file is not present, use an
    example configuration instead.
    """
    if not config_directory:
        config_directory = resolve_config_directory()
    config_file = Path(config_directory, 'astrality.yaml')

    if not config_file.is_file():
        logger.warning(
            'Configuration file not found in its expected path '
            + str(config_file) +
            '.'
        )
        config_directory = Path(__file__).parent.absolute() / 'config'
        config_file = config_directory / 'astrality.yaml'
        logger.warning(f'Using example configuration instead: "{config_file}"')
    else:
        logging.info(f'Using configuration file "{config_file}"')

    return config_directory, config_file


def dict_from_config_file(
    config_file: Path,
    with_env: Optional[bool] = True,
) -> ApplicationConfig:
    """
    Return a dictionary that reflects the contents of `config_file`.

    Environment variables are interpolated like this:
        ${env:NAME_OF_ENV_VARIABLE} -> os.environ[NAME_OF_ENV_VARIABLE]

    If with_env=True, an 'env' section is inserted into the dictionary
    containing all the environment variables.
    """

    if not config_file.is_file():  # pragma: no cover
        error_msg = f'Could not load config file "{config_file}".'
        logger.critical(error_msg)
        raise RuntimeError(error_msg)

    expanded_env_dict = generate_expanded_env_dict()
    config_string = preprocess_configuration_file(
        config_file,
        expanded_env_dict,
    )
    conf_dict = load(StringIO(config_string))

    if with_env:
        conf_dict['context/env'] = expanded_env_dict

    return conf_dict


def infer_runtime_variables_from_config(
    config_directory: Path,
    config_file: Path,
    config: ApplicationConfig,
) -> Dict[str, Dict[str, Path]]:
    """Return infered runtime variables based on config file."""

    temp_directory = Path(os.environ.get('TMPDIR', '/tmp')) / 'astrality'
    if not temp_directory.is_dir():
        os.mkdir(temp_directory)

    return {
        '_runtime': {
            'config_directory': config_directory,
            'config_file': config_file,
            'temp_directory': temp_directory,
        }
    }


def user_configuration(config_directory: Optional[Path] = None) -> ApplicationConfig:
    """
    Return Resolver object containing the users configuration.

    Create a configuration dictionary which should directly reflect the
    hierarchy of a typical `astrality.yaml` file. Users should be able to insert
    elements from their configuration directly into conky module templates. The
    mapping should be:

    ${astrality:fonts:1} -> config['fonts']['1']

    In addition, the section config['_runtime'] is inserted, which contains
    several items specifying runtime specific values. Example keys are:
    - config_directory
    - config_file
    - temp_directory
    """
    config_directory, config_file = infer_config_location(config_directory)

    config = dict_from_config_file(
        config_file,
        with_env=True,
    )
    config.update(infer_runtime_variables_from_config(
        config_directory,
        config_file,
        config,
    ))

    return config

def preprocess_configuration_file(
    conf_file: Path,
    env_dict: MutableMapping[str, str] = os.environ,
) -> str:
    """
    Interpolate environment variables and command substitutions in file.

    Interpolation syntax:
        ${name} -> os.environ[name].
        $(command) -> stdout from shell execution.
    """
    working_directory = conf_file.parent

    conf_text = ''
    with open(conf_file, 'r') as file:
        for line in file:
            conf_text += insert_environment_values(
                insert_command_substitutions(
                    content=line,
                    shell_command_working_directory=working_directory
                ),
                env_dict,
            )

    return conf_text

def insert_environment_values(
    content: str,
    env_dict: MutableMapping[str, str] = os.environ,
) -> str:
    """Replace all occurences in string: ${name} -> env_dict[name]."""

    env_dict = generate_expanded_env_dict()
    env_variable_pattern = re.compile(r'\$\{(\w+)\}')

    def expand_environment_variable(match: Match[str]) -> str:
        return env_dict[match.groups()[0]]

    return env_variable_pattern.sub(
        expand_environment_variable,
        content,
    )


def insert_command_substitutions(
    content: str,
    shell_command_working_directory: Path,
) -> str:
    """Replace all occurences in string: $(command) -> command stdout."""
    command_substitution_pattern = re.compile(r'\$\((.*)\)')

    def command_substitution(match: Match[str]) -> str:
        command = match.groups()[0]
        result = run_shell(
            command=command,
            working_directory=shell_command_working_directory,
        )
        if result == '':
            logger.error(
                f'Command substitution $({command}) returned empty stdout.'
            )
        return result

    return command_substitution_pattern.sub(
        command_substitution,
        content,
    )


def generate_expanded_env_dict() -> Dict[str, str]:
    """Return os.environ dict with all env variables expanded."""

    env_dict = {}
    for name, value in os.environ.items():
        try:
            env_dict[name] = os.path.expandvars(value)
        except ValueError as e:
            if 'invalid interpolation syntax' in str(e):
                logger.warning(f'''
                Could not use environment variable {name}={value}.
                It is too complex for expansion, using unexpanded value
                instead...
                ''')
                env_dict[name] = value
            else:
                raise

    return env_dict


def insert_into(
    context: Context,
    section: str,
    from_config_file: Path,
    from_section: str,
) -> Context:
    """
    Import section from config file into config dictionary.

    The method overwrites `config[section]` with the values from [from_section]
    defined in `from_config_file`.
    """
    logger.info(
        f'Importing context section {section} from {str(from_config_file)}',
    )

    contexts = compiler.context(dict_from_config_file(
        from_config_file,
        with_env=False,
    ))
    context[section] = contexts[from_section]

    return context


def create_config_directory(path: Optional[Path]=None, empty=False) -> Path:
    """Create application configuration directory and return its path."""
    if not path:
        path = resolve_config_directory()

    if not path.exists():
        if empty:
            logger.warning(f'Creating empty directory at "{str(path)}".')
            path.mkdir(parents=True)
        else:
            logger.warning(f'Copying over example config directory to "{str(path)}".')
            example_config_dir = Path(__file__).parent / 'config'
            copy_tree(
                src=str(example_config_dir),
                dst=str(path),
            )
    else:
        logger.warning(f'Path "{str(path)}" already exists! Delete it first.')

    return path



#!/usr/bin/env python3.6

"""The module meant to be run in order to start Astrality."""

import logging
import os
import signal
from typing import Set
import subprocess
import sys
import time

from astrality.config import user_configuration
from astrality.module import ModuleManager

logger = logging.getLogger('astrality')


def main(logging_level: str = 'INFO', test: bool = False):
    """
    Run the main process for Astrality.

    If test is set to True, then only one main loop is run as an integration
    test.
    """
    if 'ASTRALITY_LOGGING_LEVEL' in os.environ:
        # Override logging level if env variable is set
        logging_level = os.environ['ASTRALITY_LOGGING_LEVEL']

    # Set the logging level to the configured setting
    logging.basicConfig(level=logging_level)

    # Quit old astrality instances
    kill_old_astrality_processes()

    # How to quit this process
    def exit_handler(signal=None, frame=None) -> None:
        """Cleanup all temporary files and run module exit handlers."""

        # The temp directory is left alone, for two reasons:
        # 1: An empty directory uses neglible disk space.
        # 2: If this process is interrupted by another Astrality instance,
        #    we might experience race conditions when the exit handler deletes
        #    the temporary directory *after* the new Astrality instance creates it.

        logger.critical('Astrality was interrupted')
        logger.info('Cleaning up temporary files before exiting...')

        # Run all the module exit handlers
        module_manager.exit()

        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)

    # Some SIGINT signals are not properly interupted by python and converted
    # into KeyboardInterrupts, so we have to register a signal handler to
    # safeguard against such cases. This seems to be the case when conky is
    # launched as a subprocess, making it the process that receives the SIGINT
    # signal and not python. These signal handlers cause issues for \
    # NamedTemporaryFile.close() though, so they are only registrered when
    # we are not testing.
    if not test:
        signal.signal(signal.SIGINT, exit_handler)

        # Also catch kill-signkal from OS,
        # e.g. `kill $(pgrep -f "python astrality.py")`
        signal.signal(signal.SIGTERM, exit_handler)

    try:
        config = user_configuration()

        # Delay further actions if configuration says so
        startup_delay = float(config.get('settings/general', {}).get('startup_delay', '0'))
        time.sleep(startup_delay)

        module_manager = ModuleManager(config)
        module_manager.finish_tasks()

        # TODO: Implement startup delay config option

        while True:
            if module_manager.has_unfinished_tasks():
                # TODO: Log which new period which has been detected
                logger.info('New timer period detected.')
                module_manager.finish_tasks()
                logger.info(f'Period change routine finished.')

            if test:
                logger.debug('Main loop interupted since argument test=True.')
                return
            else:
                logger.info(
                    f'Waiting {module_manager.time_until_next_period()} '
                    'until next period change and ensuing update.'
                )
                time.sleep(
                    module_manager.time_until_next_period().total_seconds()
                )

    except KeyboardInterrupt:  # pragma: no cover
        exit_handler()

def other_astrality_pids() -> Set[int]:
    """Return the process ids (PIDs) of any other Astrality instances."""

    # Get all processes instanciated from this file
    result = subprocess.Popen(
        ['pgrep', '-f', 'astrality'],
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    pids = set(int(pid.strip()) for pid in result.stdout)

    # Return all the PIDs except for the PID of this process
    this_process_pid = os.getpid()
    return pids - set((this_process_pid,))


def kill_old_astrality_processes() -> None:
    """Kill all other instances of this script, to prevent duplicates."""

    pids = other_astrality_pids()
    failed_exits = 0
    for pid in pids:
        try:
            logger.info(f'Killing duplicate Astrality process with pid {pid}.')
            os.kill(pid, signal.SIGTERM)
        except OSError:
            logger.error(f'Could not kill old instance of astrality with pid {pid}.')
            logger.error('Continuing anyway...')
            failed_exits += 1

    while len(other_astrality_pids()) > failed_exits:
        # Wait for all the processes to exit properly
        time.sleep(0.2)


if __name__ == '__main__':
    main()



"""Tests for Resolver class."""
from math import inf

import pytest

from astrality.resolver import Resolver


class TestResolverClass:
    def test_initialization_of_config_class_with_no_config_parser(self):
        Resolver()

    def test_invocation_of_class_with_application_config(self, conf):
        Resolver(conf)

    def test_initialization_of_config_class_with_dict(self):
        conf_dict = {
            'key1': 'value1',
            'key2': 'value2',
            'key3': ('one', 'two', 'three'),
            'key4': {'key4-1': 'uno', 'key4-2': 'dos'}
        }
        config = Resolver(conf_dict)
        assert config == conf_dict

    def test_values_for_max_key_property(self):
        config = Resolver()
        assert config._max_key == -inf

        config['string_key'] = 1
        assert config._max_key == -inf

        config[2] = 'string_value'
        assert config._max_key == 2

        config[1] = 'string_value'
        assert config._max_key == 2

        config[3] = 'string_value'
        assert config._max_key == 3

    def test_getting_item_from_empty_config(self):
        config = Resolver()
        with pytest.raises(KeyError) as exception:
            config['empty_config_with_no_key']

    def test_accessing_existing_key(self):
        config = Resolver()
        config['some_key'] = 'some_value'
        assert config['some_key'] == 'some_value'

        config[-2] = 'some_other_value'
        assert config[-2] == 'some_other_value'

    def test_integer_index_resolution(self):
        config = Resolver()
        config['some_key'] = 'some_value'
        config[1] = 'FureCode Nerd Font'
        assert config[2] == 'FureCode Nerd Font'

    def test_integer_index_resolution_without_earlier_index_key(self):
        config = Resolver()
        config['some_key'] = 'some_value'
        with pytest.raises(KeyError) as exception:
            config[2]
        assert exception.value.args[0] == \
            'Integer index "2" is non-existent and ' \
            'had no lower index to be substituted for'

    def test_index_resolution_with_string_key(self):
        config = Resolver()
        config[2] = 'some_value'
        with pytest.raises(KeyError) as exception:
            config['test']
        assert exception.value.args[0] == 'test'

    def test_use_of_recursive_config_objects_created_by_dicts(self):
        conf_dict = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'}
        }
        config = Resolver(conf_dict)
        assert config == conf_dict
        assert config[3][2] == 'some_value'
        assert config[2] == {1: 'some_value'}
        assert config[3] == {1: 'some_value'}

        assert isinstance(config['key4'], Resolver)
        assert config['key4'] == {1: 'uno', 'key4-2': 'dos'}
        assert config['key4'][1] == 'uno'
        assert config['key4'][2] == 'uno'

    def test_getter(self):
        config = Resolver()
        assert config.get('from_empty_config') is None

        config['test'] = 'something'
        assert config.get('test') == 'something'
        assert config.get('test', '4') == 'something'

        assert config.get('non_existent_key') is None
        assert config.get('non_existent_key', '4') == '4'

    def test_items(self):
        config = Resolver()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.items()) == [('4', 'test',), ('font', 'Comic Sans',), ('5', '8',)]

    def test_keys(self):
        config = Resolver()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.keys()) == ['4', 'font', '5']

    def test_values(self):
        config = Resolver()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.values()) == ['test', 'Comic Sans', '8']

    def test_update(self):
        one_conf_dict = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
        }
        another_conf_dict = {
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'}
        }
        merged_conf_dicts = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'}
        }
        config = Resolver(one_conf_dict)
        config.update(another_conf_dict)
        assert config == merged_conf_dicts

    def test_resolver_class(self):
        resolver = Resolver()
        resolver[1] = 'firs_value'
        resolver[2] = 'second_value'
        resolver['string_key'] = 'string_value'

        assert resolver[1] == 'firs_value'
        assert resolver[2] == 'second_value'
        assert resolver[3] == 'second_value'
        assert resolver['string_key'] == 'string_value'

    def test_initializing_resolver_with_resolver(self):
        resolver1 = Resolver({'key1': 1})
        resolver2 = Resolver(resolver1)
        assert resolver1 == resolver2

    def test_updating_resolver_with_resolver(self):
        resolver1 = Resolver({'key1': 1})
        resolver2 = Resolver({'key2': 2})

        resolver1.update(resolver2)
        expected_result = Resolver({'key1': 1, 'key2': 2})
        assert resolver1 == expected_result



from datetime import datetime, timedelta

import pytest

from astrality.timer import Solar


@pytest.fixture
def solar_config():
    return {
        'type': 'solar',
        'latitude': 63.446827,
        'longitude': 10.421906,
        'elevation': 0,
    }

@pytest.fixture
def solar(solar_config):
    return Solar(solar_config)

# --- Times around dawn ---
@pytest.fixture
def dawn(solar):
    return solar.construct_astral_location().sun()['dawn']


@pytest.fixture
def before_dawn(dawn):
    delta = timedelta(minutes=-2)
    return dawn + delta


@pytest.fixture
def after_dawn(dawn):
    delta = timedelta(minutes=2)
    return dawn + delta


def test_that_night_is_correctly_identified(solar, before_dawn, freezer):
    freezer.move_to(before_dawn)
    period = solar.period()
    assert period == 'night'


def test_that_sunrise_is_correctly_identified(solar, after_dawn, freezer):
    freezer.move_to(after_dawn)
    period = solar.period()
    assert period == 'sunrise'


# --- Times around dusk ---
@pytest.fixture
def dusk(solar):
    return solar.construct_astral_location().sun()['dusk']


@pytest.fixture
def before_dusk(dusk):
    delta = timedelta(minutes=-2)
    return dusk + delta


@pytest.fixture
def after_dusk(dusk):
    delta = timedelta(minutes=2)
    return dusk + delta


def test_that_night_is_correctly_identified_after_dusk(solar, after_dusk, freezer):
    freezer.move_to(after_dusk)
    period = solar.period()
    assert period == 'night'


def test_that_sunset_is_correctly_identified_before_dusk(solar, before_dusk, freezer):
    freezer.move_to(before_dusk)
    period = solar.period()
    assert period == 'sunset'


def test_location(solar):
    location = solar.construct_astral_location()
    assert str(location) == 'CityNotImportant/RegionIsNotImportantEither, tz=UTC, lat=63.45, lon=10.42'

def test_time_left_before_new_period(solar, before_dusk, freezer):
    freezer.move_to(before_dusk)
    assert solar.time_until_next_period().total_seconds() == 120

def test_time_right_before_midnight(solar, freezer):
    """
    This function requires special handling when the UTC time is later than all
    solar periods within the same day, which is the case right before midnight.
    """

    timezone = solar.location.timezone
    before_midnight = datetime(
        year=2019,
        month=12,
        day=23,
        hour=23,
        second=59,
        microsecond=0,
    )
    freezer.move_to(before_midnight)

    # Test that the time left is within the bounds of 0 to 24 hours
    time_left = solar.time_until_next_period()
    assert 0 < time_left.total_seconds() < 60 * 60 * 24

def test_config_timer_method():
    solar_timer_application_config = {'type': 'solar'}
    solar_timer = Solar(solar_timer_application_config)
    assert solar_timer.timer_config['latitude'] == 0



import pytest

from astrality.timer import Periodic, Solar, Static, Weekday, timer_factory


@pytest.fixture
def solar_config():
    return {
        'type': 'solar',
        'longitude': 1,
        'latitude': 2,
        'elevation': 3,
    }


@pytest.fixture
def weekday_config():
    return {
        'type': 'weekday',
    }


@pytest.fixture
def periodic_config():
    return {
        'type': 'periodic',
        'seconds': 0,
        'minutes': 1,
        'hours': 2,
        'days': 3,
    }


@pytest.fixture
def static_config():
    return {
        'type': 'static',
    }


class TestTimerFactory:
    def test_timer_factory_with_solar_confi(self, solar_config):
        assert isinstance(timer_factory(solar_config), Solar)

    def test_timer_factory_with_weekday_confi(self, weekday_config):
        assert isinstance(timer_factory(weekday_config), Weekday)

    def test_timer_factory_with_periodic_confi(self, periodic_config):
        assert isinstance(timer_factory(periodic_config), Periodic)

    def test_timer_factory_with_static_confi(self, static_config):
        assert isinstance(timer_factory(static_config), Static)


@pytest.fixture
def solar_timer(solar_config):
    return Solar(solar_config)


@pytest.fixture
def weekday_timer(weekday_config):
    return Weekday(weekday_config)


@pytest.fixture
def periodic_timer(periodic_config):
    return Periodic(periodic_config)


@pytest.fixture
def static_timer(static_config):
    return Static(static_config)


class TestTimerDefaultConfiguration:
    def test_all_options_specified_of_solar_timer(
        self,
        solar_config,
        solar_timer,
    ):
        assert solar_timer.timer_config == solar_config

    def test_replacement_of_missing_timer_config_option(
        self,
        solar_config,
    ):
        solar_config.pop('elevation')
        solar_timer = Solar(solar_config)

        assert solar_timer.timer_config != solar_config
        assert solar_timer.timer_config['longitude'] == 1
        assert solar_timer.timer_config['latitude'] == 2
        assert solar_timer.timer_config['elevation'] == 0



from datetime import datetime, timedelta
import logging

import pytest

from astrality.timer import Weekday

@pytest.fixture
def weekday():
    return Weekday({'type': 'weekday'})

@pytest.fixture
def noon_friday():
    return datetime(year=2018, month=1, day=26, hour=12)


def test_weekday_periods(weekday):
    assert weekday.periods == (
        'monday',
        'tuesday',
        'wednesday',
        'thursday',
        'friday',
        'saturday',
        'sunday',
    )


def test_weekday_period(weekday, noon_friday, freezer):
    freezer.move_to(noon_friday)
    assert weekday.period() == 'friday'


def test_weekday_time_until_next_period(weekday, noon_friday, freezer):
    freezer.move_to(noon_friday)
    assert weekday.time_until_next_period() == timedelta(hours=12)


def test_using_force_period_config_option(noon_friday, freezer, caplog):
    solar_timer_application_config = {
        'type': 'weekday',
        'force_period': 'monday',
    }
    freezer.move_to(noon_friday)
    weekday_timer = Weekday(solar_timer_application_config)

    # Even though it is friday, the force option makes the timer return 'monday'
    # as the current period.
    assert weekday_timer.period() == 'monday'

    # And there are no errors logged, since 'monday' is a valid Weekday period
    assert len(caplog.record_tuples) == 0


def test_using_force_period_config_option_with_wrong_period_type(
    noon_friday,
    freezer,
    caplog,
):
    solar_timer_application_config = {
        'type': 'weekday',
        'force_period': 'Mothers_day',
    }
    freezer.move_to(noon_friday)
    weekday_timer = Weekday(solar_timer_application_config)

    # Even though it is friday, the force option makes the timer return
    # 'Mothers_day' as the current period.
    assert weekday_timer.period() == 'Mothers_day'

    # There is a warnig for using an invalid weekday period
    assert logging.WARNING == caplog.record_tuples[0][1]



"""Module for all timer classes, keeping track of certain events for modules."""

import abc
from collections import namedtuple
import logging
import time
from datetime import datetime, timedelta
from math import inf
from typing import Dict, Tuple, Union

import pytz
from astral import Location


TimerConfig = Dict[str, Union[str, int, float, None]]
logger = logging.getLogger('astrality')


class Timer(abc.ABC):
    """Class which defines different periods."""

    periods: Tuple[str, ...]
    default_timer_config: TimerConfig

    def __init__(self, timer_config: TimerConfig) -> None:
        """Initialize a period timer based on the configuration of the user."""
        self.name = self.__class__.__name__.lower()

        # Use default values for timer configuration options that are not
        # specified
        self.timer_config = self.default_timer_config.copy()
        self.timer_config.update(timer_config)

    def period(self) -> str:
        """
        Return the current determined period.

        If the timer option `force_period` is set, this value will always be
        returned instead of the correct period.
        """
        if self.timer_config.get('force_period', False):
            force_period = self.timer_config['force_period']

            if not force_period in self.periods:
                logger.warning(
                    f'[timer/{self.name}] option `force_period` set to '
                    f'{force_period}, which is not a valid period type for '
                    f'the timer type "{self.name}": {self.periods}. Still using'
                    ' the option in case it is intentional.'
                )

            return force_period  # type: ignore

        return self._period()


    @abc.abstractmethod
    def _period(self) -> str:
        """Return the current determined period."""
        pass

    @abc.abstractmethod
    def time_until_next_period(self) -> timedelta:
        """Return the time remaining until the next period in seconds."""
        pass


class Solar(Timer):
    """
    Timer subclass which keeps track of the suns position in the sky.

    It changes period after dawn, sunrise, morning, afternoon, sunset, dusk.
    """
    periods = ('sunrise', 'morning', 'afternoon', 'sunset', 'night')
    default_timer_config = {
        'type': 'solar',
        'longitude': 0,
        'latitude': 0,
        'elevation': 0,
    }

    def __init__(self, timer_config: TimerConfig) -> None:
        super().__init__(timer_config)
        self.location = self.construct_astral_location()

    def _period(self) -> str:
        now = self.now()

        if now < self.location.sun()['dawn']:
            period = 'night'
        elif now < self.location.sun()['sunrise']:
            period = 'sunrise'
        elif now < self.location.sun()['noon']:
            period = 'morning'
        elif now < self.location.sun()['sunset']:
            period = 'afternoon'
        elif now < self.location.sun()['dusk']:
            period = 'sunset'
        else:
            period = 'night'

        return period

    def time_until_next_period(self) -> timedelta:
        now = self.now()
        try:
            next_period = min(
                utc_time
                for utc_time
                in self.location.sun().values()
                if now < utc_time
            )
        except ValueError as exception:
            if str(exception) == 'min() arg is an empty sequence':
                # None of the solar periods this current day are in the future,
                # so we need to compare with solar periods tomorrow instead.
                tomorrow = now + timedelta(days=1, seconds=-1)
                next_period = min(
                    utc_time
                    for utc_time
                    in self.location.sun(tomorrow).values()
                    if now < utc_time
                )
            else:
                raise RuntimeError('Could not find the time of the next period')

        return next_period - now

    def now(self) -> datetime:
        """Return the current UTC time."""
        timezone = pytz.timezone('UTC')
        return timezone.localize(datetime.utcnow())

    def construct_astral_location(
        self,
    ) -> Location:
        # Initialize a custom location for astral, as it doesn't necessarily include
        # your current city of residence
        location = Location()

        # These two doesn't really matter
        location.name = 'CityNotImportant'
        location.region = 'RegionIsNotImportantEither'

        # But these are important, and should be provided by the user
        location.latitude = self.timer_config['latitude']
        location.longitude = self.timer_config['longitude']
        location.elevation = self.timer_config['elevation']
        location.timezone = 'UTC'

        return location


class Weekday(Timer):
    """Timer subclass which keeps track of the weekdays."""

    periods = (
        'monday',
        'tuesday',
        'wednesday',
        'thursday',
        'friday',
        'saturday',
        'sunday',
    )
    default_timer_config = {'type': 'weekday'}

    weekdays = dict(zip(range(0,7), periods))

    @classmethod
    def _period(cls) -> str:
        """Return the current determined period."""
        return cls.weekdays[datetime.today().weekday()]

    def time_until_next_period(self) -> timedelta:
        """Return the time remaining until the next period in seconds."""
        tomorrow = datetime.today() + timedelta(days=1)
        tomorrow = tomorrow.replace(hour=0, minute=0)
        return tomorrow - datetime.now()


class Periodic(Timer):
    """Constant frequency Timer subclass."""

    class Periods(tuple):
        def __contains__(self, item) -> bool:
            try:
                if float(item) - int(item) != 0:
                    return False
                else:
                    return 0 <= int(item) < inf
            except ValueError:
                return False

    periods = Periods()
    default_timer_config = {
        'type': 'periodic',
        'seconds': 0,
        'minutes': 0,
        'hours': 0,
        'days': 0,
    }

    def __init__(self, timer_config: TimerConfig) -> None:
        """Initialize a constant frequency timer."""

        super().__init__(timer_config)

        # Period specified by the user
        self.timedelta = timedelta(  # type: ignore
            seconds=self.timer_config['seconds'],
            minutes=self.timer_config['minutes'],
            hours=self.timer_config['hours'],
            days=self.timer_config['days'],
        )

        if self.timedelta.total_seconds() == 0.0:
            # If no period is specified by the user, then 1 hour is used
            self.timedelta = timedelta(hours=1)

        self.initialization_time = datetime.now()

    def _period(self) -> str:
        return str(int(
            (datetime.now() - self.initialization_time) / self.timedelta
        ))

    def time_until_next_period(self) -> timedelta:
        """Return the time remaining until the next period in seconds."""
        return self.timedelta - (datetime.now() - self.initialization_time) % self.timedelta


WorkDay = namedtuple('WorkDay', ('start', 'end',))

class TimeOfDay(Timer):
    """
    Timer which keeps track of time intervals each day of the week.

    The variable names used in the implementation, assumes that the 'on'
    periods are referring to worktime.
    """

    periods = ('on', 'off',)
    default_timer_config = {
        'type': 'time_of_day',
        'monday': '09:00-17:00',
        'tuesday': '09:00-17:00',
        'wednesday': '09:00-17:00',
        'thursday': '09:00-17:00',
        'friday': '09:00-17:00',
        'saturday': '',
        'sunday': '',
    }

    def __init__(self, timer_config: TimerConfig) -> None:
        super().__init__(timer_config)
        self.weekdays = (
            'monday',
            'tuesday',
            'wednesday',
            'thursday',
            'friday',
            'saturday',
            'sunday',
        )
        self.workdays: Dict[str, WorkDay] = {}

        for weekday_num, weekday_name in enumerate(self.weekdays):
            work_period = self.timer_config[weekday_name]
            if not work_period:
                continue

            work_start, work_end = work_period.split('-')  # type: ignore
            workday = WorkDay(
                start=time.strptime(work_start, '%H:%M'),
                end=time.strptime(work_end, '%H:%M'),
            )
            self.workdays[weekday_name] = workday

    def _period(self) -> str:
        """Return the current determined period."""
        weekday_name = Weekday({'type': 'weekday'}).period()
        if weekday_name not in self.workdays:
            return 'off'
        else:
            now = datetime.now()
            now_hour = now.hour
            now_minute = now.minute

            after_work_start = \
                now_hour >= self.workdays[weekday_name].start.tm_hour and \
                now_minute >= self.workdays[weekday_name].start.tm_min

            before_work_end = \
                now_hour <= self.workdays[weekday_name].end.tm_hour and \
                now_minute <= self.workdays[weekday_name].end.tm_min

            if after_work_start and before_work_end:
                return 'on'
            else:
                return 'off'


    def time_until_next_period(self) -> timedelta:
        """Return the time remaining until the next period in seconds."""
        weekday_name = Weekday({'type': 'weekday'}).period()

        now = datetime.now()
        now_hour = now.hour
        now_minute = now.minute

        if self._period() == 'on':
            # We are within work hours, and it is easy to find the work end for
            # the same day.
            return timedelta(
                hours=self.workdays[weekday_name].end.tm_hour - now_hour,
                minutes=self.workdays[weekday_name].end.tm_min - now_minute + 1,
            )

        else:
            # TODO: This is bad code, and should be done more cleanly in the
            #       future. But it works, and that is good enough for now.

            # The current zero-indexed weekday
            weekday_num = self.weekdays.index(weekday_name)

            # Get the zero indexed workdays
            workday_nums = tuple(
                self.weekdays.index(workday)
                for workday
                in self.workdays
            )

            # Retrieve all zero-indexed workdays that are later in this week
            next_workday_nums = tuple(
                workday_num
                for workday_num
                in workday_nums
                if workday_num > weekday_num
            )

            if len(next_workday_nums) > 0:
                days_until_next_workday = next_workday_nums[0] - weekday_num
                next_workday = self.workdays[self.weekdays[next_workday_nums[0]]]
            else:
                # There are no workdays later in this week, so we need to
                # get the *first* workday from next week instead.
                min_workday_num = min(workday_nums)
                days_until_next_workday = 7 + min_workday_num - weekday_num
                next_workday = self.workdays[self.weekdays[min_workday_num]]

            return timedelta(
                days=days_until_next_workday,
                hours=next_workday.start.tm_hour - now_hour,
                minutes=next_workday.start.tm_min - now_minute + 1,
            )


class Static(Timer):
    """Timer subclass which never changes period."""

    periods = ('static',)
    default_timer_config = {'type': 'static'}

    def _period(self) -> str:
        """Static timer asways returns the period 'static'."""

        return 'static'

    def time_until_next_period(self) -> timedelta:
        """Returns a 100 year timedelta as an infinite approximation."""

        return timedelta(days=36500)


TIMERS = {
        'solar': Solar,
        'weekday': Weekday,
        'periodic': Periodic,
        'static': Static,
}

def timer_factory(timer_config: Dict[str, Union[str, int]]) -> Timer:
    timer_type = timer_config['type']
    return TIMERS[timer_type](timer_config)  # type: ignore



"""General utility functions which are used across the application."""

import logging
import subprocess
from pathlib import Path
from typing import Union

logger = logging.getLogger('astrality')


def run_shell(
    command: str,
    timeout: Union[int, float] = 2,
    fallback: str = '',
    working_directory: Path = Path.home(),
) -> str:
    """Return the standard output of a shell command."""

    process = subprocess.Popen(
        command,
        cwd=working_directory,
        shell=True,
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        process.wait(timeout=timeout)

        for error_line in process.stderr:
            logger.error(str(error_line))

        if process.returncode != 0:
            logger.error(
                f'Command "{command}" exited with non-zero return code: ' +
                str(process.returncode)
            )
            return fallback
        else:
            stdout = process.communicate()[0]
            logger.info(stdout)
            return stdout.replace('\n', '')

    except subprocess.TimeoutExpired:
        logger.warning(
            f'The command "{command}" used more than 2 seconds in order to '
            'finish. The exit code can not be verified. This might be '
            'intentional for background processes and daemons.'
        )
        return fallback
